/*  This file is part of FediBlock-backend.
 *  Copyright © 2020 tastytea <tastytea@tastytea.de>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, version 3.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "rss.hpp"
#include "cgi.hpp"
#include "files.hpp"
#include "git.hpp"
#include "time.hpp"

#include <fmt/format.h>

#include <algorithm>
#include <chrono>
#include <cstdint>
#include <cstdlib>
#include <exception>
#include <iostream>
#include <ostream>
#include <stdexcept>
#include <string>
#include <string_view>
#include <vector>

namespace FediBlock::rss
{

using fmt::format;
using std::find;
using std::getenv;
using std::none_of;
using std::ostream;
using std::string;
using std::string_view;
using std::uint8_t;
using std::chrono::system_clock;

void write_line(ostream &out, const uint8_t spaces, const string_view rsstag,
                const string_view value)
{
    // clang-format off
    const string endtag{[&rsstag]
    {
        // If there is a space in the rsstag, use only the part up until the
        // space for the ending RSS tag.
        const size_t pos = rsstag.find(' ');
        if (pos == string_view::npos)
        {
            return rsstag;
        }
        return rsstag.substr(0, pos);
    }()};
    // clang-format off

    out << string(spaces, ' ');
    out << '<' << rsstag << '>' << value << "</" << endtag << ">\n";
}

void write_rss(ostream &out, const vector<entry_type> &entries,
               const vector<string> &tags)
{
    string selfurl;
    char *env{getenv("HTTPS")};
    if (env == nullptr || string(env) != "on")
    {
        selfurl = "http://";
    }
    else
    {
        selfurl = "https://";
    }
    env = getenv("SERVER_NAME");
    if (env != nullptr)
    {
        selfurl += env;
    }
    const string baseurl{selfurl + "/blocklist/"};
    env = getenv("REQUEST_URI");
    if (env != nullptr)
    {
        selfurl += env;
    }

    constexpr string_view rss_time_format{"%a, %d %b %Y %T %z"};

    out << R"(<rss version="2.0" xmlns:atom="http://www.w3.org/2005/Atom">
  <channel>
)";
    out << R"(    <atom:link href=")" + selfurl
        << R"(" rel="self" type="application/rss+xml"/>)" << '\n';
    string title{"FediBlock: "};
    string description{"The newest FediBlock entries. "};
    if (tags.empty())
    {
        title += "All tags";
        description += "All tags";
    }
    else
    {
        description += "Tags: ";
    }
    for (const auto &tag : tags)
    {
        if (tag != *tags.begin())
        {
            title += ", ";
            description += ", ";
        }
        title += tag;
        description += tag;
    }
    description += ".";
    write_line(out, 4, "title", title);
    write_line(out, 4, "link", baseurl);
    write_line(out, 4, "description", description);
    write_line(out, 4, "lastBuildDate",
               time::to_string(system_clock::now(), rss_time_format));

    for (const auto &entry : entries)
    {
        // If tags are specified and none of them match, go to the next entry.
        if (!tags.empty())
        {
            // clang-format off
            if (none_of(entry.tags.begin(), entry.tags.end(),
                        [&tags](const auto &tag)
                        {
                            return find(tags.begin(), tags.end(), tag)
                                   != tags.end();
                        }))
            {
                continue;
            }
            // clang-format on
        }

        out << "    <item>\n";
        write_line(out, 6, "title", entry.instance);
        write_line(out, 6, "guid isPermaLink=\"false\"",
                   "FediBlock: " + entry.report_time + " " + entry.instance);
        write_line(out, 6, "pubDate",
                   time::to_string(entry.report_time, rss_time_format));
        write_line(out, 6, "link",
                   format("{:s}#{:s}", baseurl, entry.instance));

        string item_description{format("<p>{:s}</p><p><strong>Tags:</strong> ",
                                       cgi::text2html(entry.description))};
        for (const auto &tag : entry.tags)
        {
            if (tag != *entry.tags.begin())
            {
                item_description += ", ";
            }
            item_description += tag;
        }
        item_description += "</p><strong>Receipts:</strong><ul>";
        for (const auto &receipt : entry.receipts)
        {
            item_description +=
                format("<li><a href=\"{0:s}\">{0:s}</a></li>", receipt);
        }
        item_description += "</ul>";
        if (!entry.screenshot_filepath.empty())
        {
            item_description += "<p><strong>Screenshot:</strong><br>";
            item_description +=
                format("<a href=\"{0:s}{1:s}\"><img "
                       "src=\"{0:s}{1:s}\" height=\"200\"></a></p>\n",
                       baseurl, entry.screenshot_filepath);
        }
        write_line(out, 6, "description",
                   format("<![CDATA[{:s}]]>", item_description));

        out << "    </item>\n";
    }

    out << "  </channel>\n</rss>\n";
}

} // namespace FediBlock::rss

using std::cerr;
using std::cout;
using std::exception;
using std::runtime_error;
using std::vector;

using namespace FediBlock;
using namespace FediBlock::rss;

int main(int argc, char *argv[])
{
    const vector<string_view> args(argv, argv + argc);

    git::init(true);

    bool remove_lockfile{false};
    try
    {
        try
        {
            git::update_cached_repo();
            remove_lockfile = true;
        }
        catch (const runtime_error &)
        {
            // Ignore, use old version of repo.
        }
        const auto entries{files::read_json_files(true)};
        cout << "Content-Type: application/rss+xml\r\n\r\n";
        write_rss(cout, entries, cgi::get_tags());
    }
    catch (const exception &e)
    {
        cerr << "Error: " << e.what() << '\n';
    }

    git::cleanup(remove_lockfile);

    return 0;
}
